/*
 *
 * COPYRIGHT NVIDIA CORPORATION 2007. ALL RIGHTS RESERVED.  
 * BY ACCESSING OR USING THIS SOFTWARE, YOU AGREE TO: 
 *
 *  1) ACKNOWLEDGE NVIDIA'S EXCLUSIVE OWNERSHIP OF ALL RIGHTS 
 *     IN AND TO THE SOFTWARE; 
 *
 *  2) NOT MAKE OR DISTRIBUTE COPIES OF THE SOFTWARE WITHOUT 
 *     INCLUDING THIS NOTICE AND AGREEMENT; 
 *
 *  3) ACKNOWLEDGE THAT TO THE MAXIMUM EXTENT PERMITTED BY 
 *     APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS* AND 
 *     THAT NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, 
 *     EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED
 *     TO, IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS 
 *     FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY 
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES 
 * WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS 
 * OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
 * INFORMATION, OR ANY OTHER PECUNIARY LOSS), INCLUDING ATTORNEYS'
 * FEES, RELATING TO THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 */

#include "TerrainDemo.h"

#include <NRender/NRWindow.h>
#include <NRender/NRVertexBuffer.h>

#include <NHelper/NVSliderManager.h>
#include <NVDemo/NVScene.h>
#include <NVDemo/NVTexture.h>
#include <NVDemo/NVMenu.h>
#include <NVDemo/NVMesh.h>

#include "helpers/NVAssist.h"

#include <commctrl.h>  // for TBM_SETPOS
#include <shlobj.h> // for finding the My Pictures folder

extern TerrainDemo* g_instance;

int AddPtsSwizzled( std::vector<NPnt2f>& vec, int x0, int x1, int y0, int y1, int spanX, int spanY)
{
  int count = 0;
  int xm = (x0+x1)/2;
  int ym = (y0+y1)/2;
  if (x0 == x1) {
    if (y0 == y1) {  // 1x1
      float x = x0 / (float)(spanX-1);   //   /63
      float y = y0 / (float)(spanY-1);   //   /63
      vec.push_back(NPnt2f(x,y));  // stores the lower-left coords of the cell
	  count++;
    }
    else {   // 1xN (WxH)
      count += AddPtsSwizzled(vec, x0, x1, y0  , ym, spanX, spanY);
      count += AddPtsSwizzled(vec, x0, x1, ym+1, y1, spanX, spanY);
    }
  }
  else if (y0 == y1) {   // Nx1 (WxH)
    count += AddPtsSwizzled(vec, x0  , xm, y0, y1, spanX, spanY);
    count += AddPtsSwizzled(vec, xm+1, x1, y0, y1, spanX, spanY);
  }
  else {  // 2x2 or greater
    count += AddPtsSwizzled(vec, x0  , xm, y0  , ym, spanX, spanY);
    count += AddPtsSwizzled(vec, xm+1, x1, y0  , ym, spanX, spanY);
    count += AddPtsSwizzled(vec, x0  , xm, ym+1, y1, spanX, spanY);
    count += AddPtsSwizzled(vec, xm+1, x1, ym+1, y1, spanX, spanY);
  }
  return count;
}

//----------------------------------------
//
//----------------------------------------
void TerrainDemo::Consume(NVScene* scene, const std::string& Filename )
{
  //if (m_consolidatedNMA==true)  // do not try to consume any file, if
    //return;                     // we are using a consolidated file

  NVFileReader fileReader;
  // 
  // load and merge hand edited file
  //=====================================================
  {
    NVScene *customScene = NULL;
    if (!fileReader.ReadModel(Filename, customScene)) 
    {
      NOutputString("NVDemoTest: Error reading custom file: ");
      NOutputString(Filename.c_str());
      NOutputString("'\n");
    }
    if (customScene) 
    {
      NOutputString("Merging scene '");
      NOutputString(Filename.c_str());
      NOutputString("'\n");
      scene->Consume(customScene);
    }     
  }	 
}

//----------------------------------------
//
//----------------------------------------
void TerrainDemo::PreSetSceneInitDemo(NVScene* scene)
{
  // SET UP PREPROCESSOR DEFINITIONS FOR SHADERS HERE:
  char buf[64];
  sprintf(buf, "%d", m_nUberQuality);
  m_viewer->GetRenderer()->AddShaderPreprocDefinition("UBER_QUALITY", buf);

  InitMenuAndSliders(scene);



  // cache pointers:
  m_build_density_pass          = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("build_density_pass"    ));
  m_list_nonempty_cells_pass    = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("list_nonempty_cells"   ));
  m_list_verts_to_generate_pass = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("list_verts_to_generate"));
  m_splat_vertex_IDs_pass       = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("splat_vertex_IDs"      ));
  m_gen_vertices_pass           = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("gen_vertices"          ));
  m_gen_indices_pass            = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("gen_indices"           ));

  m_build_density_shape          = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("build_density_shape"         ));
  m_list_nonempty_cells_shape    = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("list_nonempty_cells_shape"   ));
  m_list_verts_to_generate_shape = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("list_verts_to_generate_shape"));
  m_splat_vertex_IDs_shape       = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("splat_vertex_IDs_shape"      ));
  m_gen_vertices_shape           = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("gen_vertices_shape"          ));
  m_gen_indices_shape            = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("gen_indices_shape"           ));

  m_lod_CBs[0] = dynamic_cast<NVConstantBuffer*>(NVNAMESPACE->GetByName("g_LodCB_hi"));
  m_lod_CBs[1] = dynamic_cast<NVConstantBuffer*>(NVNAMESPACE->GetByName("g_LodCB_med"));
  m_lod_CBs[2] = dynamic_cast<NVConstantBuffer*>(NVNAMESPACE->GetByName("g_LodCB_lo"));
  m_ChunkCB    = dynamic_cast<NVConstantBuffer*>(NVNAMESPACE->GetByName("g_ChunkCB"));
  
  // HERE WE LIST ALL SHADER PARAMS 
  // THAT REFERENCE THE "LodCB" CONSTANT BUFFER.
  // (SHADERS THAT GET USED IN CHUNK GENERATION ONLY, THOUGH.)
  NVConstantBufferRef* ref;
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("build_density_shape"      )->FindParam(NR_VERTEX_SHADER  , "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("build_density_shape"      )->FindParam(NR_FRAGMENT_SHADER, "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("list_nonempty_cells_shape")->FindParam(NR_VERTEX_SHADER  , "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("splat_vertex_IDs_shape"   )->FindParam(NR_VERTEX_SHADER  , "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("gen_vertices_shape"       )->FindParam(NR_VERTEX_SHADER  , "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  ref = dynamic_cast<NVConstantBufferRef*>(FindShape("gen_indices_shape"        )->FindParam(NR_GEOMETRY_SHADER, "LodCB")); assert(ref); m_lod_CB_refs.push_back(ref); 
  
  for (int i=0; i<NUM_VB; i++) {
    char buf[256];
    
    sprintf(buf, "vb_%03d", i);
    NVStreamOutGeometry* g = dynamic_cast<NVStreamOutGeometry*>(NVNAMESPACE->GetByName(buf)); assert(g);
    m_vb_list[i] = g;
    
    sprintf(buf, "ib_%03d", i);
    g = dynamic_cast<NVStreamOutGeometry*>(NVNAMESPACE->GetByName(buf)); assert(g);
    m_ib_list[i] = g;
  }


  // set VoxelDim in CB's
  std::string cbNames[3] = { "g_LodCB_hi", "g_LodCB_med", "g_LodCB_lo" };
  for (int lod=0; lod<3; lod++) {
    NVConstantBuffer* cb = dynamic_cast<NVConstantBuffer*>(NVNAMESPACE->GetByName(cbNames[lod].c_str())); 
    assert(cb);
    NVFloatParam* f[32];
    f[0] = dynamic_cast<NVFloatParam*>(cb->FindParam("VoxelDim")); 
    f[1] = dynamic_cast<NVFloatParam*>(cb->FindParam("VoxelDimMinusOne")); 
    f[2] = dynamic_cast<NVFloatParam*>(cb->FindParam("wsVoxelSize")); 
    f[3] = dynamic_cast<NVFloatParam*>(cb->FindParam("wsChunkSize")); 
    f[4] = dynamic_cast<NVFloatParam*>(cb->FindParam("InvVoxelDim")); 
    f[5] = dynamic_cast<NVFloatParam*>(cb->FindParam("InvVoxelDimMinusOne")); 
    f[6] = dynamic_cast<NVFloatParam*>(cb->FindParam("Margin")); 
    f[7] = dynamic_cast<NVFloatParam*>(cb->FindParam("VoxelDimPlusMargins")); 
    f[8] = dynamic_cast<NVFloatParam*>(cb->FindParam("VoxelDimPlusMarginsMinusOne")); 
    f[9] = dynamic_cast<NVFloatParam*>(cb->FindParam("InvVoxelDimPlusMargins")); 
    f[10]= dynamic_cast<NVFloatParam*>(cb->FindParam("InvVoxelDimPlusMarginsMinusOne")); 
    f[0]->SetValue( NPnt4f( (float)m_nVoxelDim, 0, 0, 0 ));
    f[1]->SetValue( NPnt4f( (float)m_nVoxelDim-1, 0, 0, 0 ));
    f[2]->SetValue( NPnt4f( (float)((double)m_wsChunkSize[lod]/(double)(m_nVoxelDim-1)), 0, 0, 0 ));
    f[3]->SetValue( NPnt4f( (float)m_wsChunkSize[lod], 0, 0, 0 ));
    f[4]->SetValue( NPnt4f( (float)(1.0/(double)(m_nVoxelDim  )), 0, 0, 0 ));
    f[5]->SetValue( NPnt4f( (float)(1.0/(double)(m_nVoxelDim-1)), 0, 0, 0 ));
    f[6]->SetValue( NPnt4f( (float)((double)(m_nMargin)), 0, 0, 0 ));
    f[7]->SetValue( NPnt4f( (float)((double)(m_nVoxelDimPlusMargins  )), 0, 0, 0 ));
    f[8]->SetValue( NPnt4f( (float)((double)(m_nVoxelDimPlusMargins-1)), 0, 0, 0 ));
    f[9]->SetValue( NPnt4f( (float)(1.0/(double)(m_nVoxelDimPlusMargins  )), 0, 0, 0 ));
    f[10]->SetValue(NPnt4f( (float)(1.0/(double)(m_nVoxelDimPlusMargins-1)), 0, 0, 0 ));
  }

  NVProc3DTexture* tex;
  tex = dynamic_cast<NVProc3DTexture*>(NVNAMESPACE->GetByName("density_vol" )); assert(tex);
  tex->SetDimensions( NVec3i(m_nVoxelDimPlusMargins, m_nVoxelDimPlusMargins, m_nVoxelDimPlusMargins) );
  tex = dynamic_cast<NVProc3DTexture*>(NVNAMESPACE->GetByName("VertexIDVolume" )); assert(tex);
  tex->SetDimensions( NVec3i(m_nVoxelDim*3, m_nVoxelDim, m_nVoxelDim) ); //195x65x65

  // set up thick_slice_geo
  if (1) 
  {
    NVMesh* g = dynamic_cast<NVMesh*>(NVNAMESPACE->GetByName("dummy_cells"));
    assert(g);
    
    // set up vertices
    NVPnt2fArrayNode* d  = dynamic_cast<NVPnt2fArrayNode*>(g->GetData()[0]); assert(d);
    NVPnt2fArrayNode* d2 = dynamic_cast<NVPnt2fArrayNode*>(g->GetData()[1]); assert(d2);
    assert(d ->GetChannelName() == "a2v.uv_write");
    assert(d2->GetChannelName() == "a2v.uv_read");
    std::vector<NPnt2f> vec;
    std::vector<NPnt2f> vec2;
    AddPtsSwizzled(vec , 0, m_nVoxelDim-2, 0, m_nVoxelDim-2, m_nVoxelDim, m_nVoxelDim);
    AddPtsSwizzled(vec2, 0+m_nMargin, m_nVoxelDim-2+m_nMargin, 0+m_nMargin, m_nVoxelDim-2+m_nMargin, m_nVoxelDim+m_nMargin*2, m_nVoxelDim+m_nMargin*2);
    d ->SetValue( vec );
    d2->SetValue( vec2 );

    // set up indices
    NVVertexIndexNonIndexed* idx = dynamic_cast<NVVertexIndexNonIndexed*>(g->GetIndices()[0]);
    assert(idx);
    idx->SetNumVertices( vec.size() );    // one for each CELL
  }

  if (1) 
  {
    NVMesh* g = dynamic_cast<NVMesh*>(NVNAMESPACE->GetByName("dummy_corners"));
    assert(g);
    
    // set up vertices
    NVPnt2fArrayNode* d  = dynamic_cast<NVPnt2fArrayNode*>(g->GetData()[0]); assert(d);
    NVPnt2fArrayNode* d2 = dynamic_cast<NVPnt2fArrayNode*>(g->GetData()[1]); assert(d2);
    assert(d ->GetChannelName() == "a2v.uv_write");
    assert(d2->GetChannelName() == "a2v.uv_read");
    std::vector<NPnt2f> vec;
    std::vector<NPnt2f> vec2;
    AddPtsSwizzled(vec , 0,           m_nVoxelDim-2+1          , 0          , m_nVoxelDim-2+1          , m_nVoxelDim,             m_nVoxelDim            );
    AddPtsSwizzled(vec2, 0+m_nMargin, m_nVoxelDim-2+m_nMargin+1, 0+m_nMargin, m_nVoxelDim-2+m_nMargin+1, m_nVoxelDim+m_nMargin*2, m_nVoxelDim+m_nMargin*2);
    d ->SetValue( vec );
    d2->SetValue( vec2 );

    // set up indices
    NVVertexIndexNonIndexed* idx = dynamic_cast<NVVertexIndexNonIndexed*>(g->GetIndices()[0]);
    assert(idx);
    idx->SetNumVertices( vec.size() );    // one for each CELL
  }

  // set NumInstances on the shapes that render into a 3D texture:
  m_build_density_shape->SetNumInstances( m_nVoxelDimPlusMargins );  
  m_list_nonempty_cells_shape->SetNumInstances( m_nVoxelDim );

}

void TerrainDemo::InitLogo()
{
    NVShape *logo_s = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("nvLogo_quadProxy_shape"));
    if(!logo_s)
    {
        return;
    }
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(logo_s->GetVertexParam("scale"));
    if(!param)
    {
        return;
    }
    NVFloatParam *offsetParam = dynamic_cast<NVFloatParam*>(logo_s->GetVertexParam("offset"));
    if(!offsetParam)
    {
        return;
    }

    int dx = 200; //lazy
    int dy = 156;
    float logoYPercent = dy/(float)(m_winDimensions.y);
    float logoXPercent = dx/(float)(m_winDimensions.x);
    param->SetValue(NPnt4f(logoXPercent, logoYPercent, 0, 0));

    float positionOffset = (float)(m_winDimensions.y - m_renDimensions.y);
    positionOffset = positionOffset/(float)(m_winDimensions.y);

    NPnt4f position(1.0f - logoXPercent /*- 0.015f*/, -1.0f + logoYPercent /*- 0.032f*/, 0, 0);
    position.y += positionOffset;
    offsetParam->SetValue(position);

    // add it to the main shapegroup
    if (m_bShowLogo) {
      NVShapeGroup *logoGroup  = dynamic_cast<NVShapeGroup*>(NVNAMESPACE->GetByName("nvLogo_group"));
      NVRenderPass *primaryPass = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("nvPrimaryRenderPass"));
      assert(primaryPass);
      assert(logoGroup);
      RemoveShapeFromGroup(logo_s, logoGroup);
      primaryPass->AppendShapes(logo_s);    
    }
}

void TerrainDemo::InitClickHelp()
{
    NVShape *pressF1_s = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("clickhelp_shape"));
    if(!pressF1_s)
    {
        return;
    }
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(pressF1_s->GetVertexParam("scale"));
    if(!param)
    {
        return;
    }
    NVFloatParam *offsetParam = dynamic_cast<NVFloatParam*>(pressF1_s->GetVertexParam("offset"));
    if(!offsetParam)
    {
        return;
    }

    float logoXPercent = 116 / (float)m_winDimensions.x;
    float logoYPercent =  62 / (float)m_winDimensions.y;
    param->SetValue(NPnt4f(logoXPercent, logoYPercent, 0, 0));

    float positionOffset = (float)(m_winDimensions.y - m_renDimensions.y);
    positionOffset = positionOffset/(float)(m_winDimensions.y);

    NPnt4f position(-1.0f + logoXPercent + 0.005f, -1.0f + logoYPercent - 0.005f, 0, 0);
    position.y += positionOffset;
    offsetParam->SetValue(position);

    // add it to the main shapegroup
    if (m_bShowLogo) {
      NVShapeGroup *logoGroup  = dynamic_cast<NVShapeGroup*>(NVNAMESPACE->GetByName("nvLogo_group"));
      NVRenderPass *primaryPass = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName("nvPrimaryRenderPass"));
      assert(primaryPass);
      assert(logoGroup);
      RemoveShapeFromGroup(pressF1_s, logoGroup);
      primaryPass->AppendShapes(pressF1_s);    
    }
}


#include <fstream>
#include <iostream>

DemoSoundFmod* TerrainDemo::AddLoopingSound(char* filename, float vol)
{
  DemoSoundFmod* p = new DemoSoundFmod();
  
  // check that file exists, b/c if it doesn't, FMod *exits the process*!
  DWORD att = ::GetFileAttributes(filename);
  if (att == INVALID_FILE_ATTRIBUTES)
    return NULL;

  p->Create(filename, true, vol);
  //if (vol > 0)
    p->Play();
  m_sounds.push_back(p);
  return p;
}

//----------------------------------------
//
//----------------------------------------
void TerrainDemo::PostSetSceneInitDemo(NVScene* scene)
{
  InitLogo();
  InitClickHelp();

  Randomize();

  NVShape* clear_vertex_IDs_shape = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("clear_vertex_IDs_shape"));
  assert(clear_vertex_IDs_shape);
  clear_vertex_IDs_shape->SetNumInstances( m_nVoxelDim );

  // init audio
  DemoSoundFmod::OpenSoundSystem();
  //m_pSoundBkg = AddLoopingSound("..\\audio\\bkg1.ogg", 1.0f);
  //m_pSoundEffect = AddLoopingSound("..\\audio\\effect.mp3", 1.0f);
  // etc.


}

// This is called BEFORE SetScene() to make sure that the targets are
// of appropriate size and sample depths before allocation
void 
TerrainDemo::ApplyWinDimensions(NVScene *scene){

  NVProc2DTexture *target = NULL;

  #define GLOW_SIZE_DIV 2

  NVec2i frameBufferDims = NVec2i( RoundUpTo16((int)(m_renDimensions.x * m_fSuperSampling)), 
                                   RoundUpTo4 ((int)(m_renDimensions.y * m_fSuperSampling)) );

  // primary color + depth:
  target = dynamic_cast<NVProc2DTexture*>(NVNAMESPACE->GetByName("sceneColorTarget"));
  target->SetDimensions( frameBufferDims );
  target->SetNumSamples( 1 );
  target = dynamic_cast<NVProc2DTexture*>(NVNAMESPACE->GetByName("sceneDepthTarget"));
  target->SetDimensions( frameBufferDims );
  target->SetNumSamples( 1 );

  // post-processing textures:
  target = dynamic_cast<NVProc2DTexture*>(NVNAMESPACE->GetByName("postproc_A"));
  target->SetDimensions(m_renDimensions / GLOW_SIZE_DIV);
  target->SetNumSamples(1);
  
  target = dynamic_cast<NVProc2DTexture*>(NVNAMESPACE->GetByName("postproc_B"));
  target->SetDimensions(m_renDimensions / GLOW_SIZE_DIV);
  target->SetNumSamples(1);



  // camera aspect ratio should be determined by window size
  NVPerspectiveCamera *camera = dynamic_cast<NVPerspectiveCamera*>(NVNAMESPACE->GetByName("mainCamShape"));
  if (camera != NULL){
    camera->SetUseScreenAspectRatio(false);
    float renderedRatio = float(m_renDimensions.x) / float(m_renDimensions.y);
    float winRatio      = float(m_winDimensions.x) / float(m_winDimensions.y);
    float aspectRatio   = renderedRatio;
    camera->SetAspectRatio(aspectRatio);
  }


}

// This is called AFTER the slider values are read and applied to ensure
// that we get the appropriate aspect ratio in the blitting scaling passes
void 
TerrainDemo::UpdateWinDimensions(NVScene *scene){
	
  // how much do we have to squish the final quad to get a 16:9 output?
  m_quadScaleFactor = m_pixelCorrection * (float)m_renDimensions.y / (float)m_winDimensions.y;
  NPnt4f scale = NPnt4f(1.0f, m_quadScaleFactor, 1.0f, m_renDimensions.y/(float)m_renDimensions.x);

  // Scaling render target quad for final display
  // For now we only have to set this once.
  // Perhaps we will have to set it every frame some day?
  char szQuads[][64] = { "quadProxy_shape_nvPrimaryRenderPass", 
                         //"quadProxy_shape_nvMotionBlurPass",
                         //"quadProxy_shape_nvMotionBlurPass_2",
  };
  for (int i=0; i<sizeof(szQuads)/sizeof(szQuads[0]); i++) {
    NVShape *compositeQuadShape = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName(szQuads[i]));
    assert(compositeQuadShape);
    NVFloatParam *cp;
    cp = dynamic_cast<NVFloatParam*>(compositeQuadShape->GetVertexParam("scale"));
    assert(cp);
    cp->SetValue(scale);
    cp = dynamic_cast<NVFloatParam*>(compositeQuadShape->GetFragmentParam("scale"));
    assert(cp);
    cp->SetValue(scale);
  }

  //NVShape* hs = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName("helpscr_shape")); assert(hs);
  //cp = dynamic_cast<NVFloatParam*>(hs->GetVertexParam("scale")); assert(cp);
  //cp->SetValue(scale);
}



//----------------------------------------
//
//----------------------------------------
void TerrainDemo::InitDemo(NVScene* scene)
{
  ResolveReferencesRecur(scene);
  
  m_viewer->m_preDrawListeners.push_back( this );
  m_viewer->m_renderPassListeners.push_back( this );

  PreSetSceneInitDemo(scene);  

  // create parameter sliders
  m_sliderManager = new NVSliderManager(m_viewer, scene);
  m_sliderManager->BuildParamsMenus(scene);

  // Build the menu selector to scroll through menus.
  // The menu should start hidden
  NVStaticMenu *sliderMenu = m_sliderManager->GetSliderMenu();
  m_menuSelector = new NVMenuSelector(sliderMenu);
  sliderMenu->SetVisible( false );
  
  // read slider file and set values in scene
  m_sliderManager->LoadValues(m_sliderFile);

  // Propogate our arguments to the people who care
  ApplyWinDimensions(scene);

  // set the scene on the viewer
  m_viewer->SetScene(scene);  
  
  UpdateWinDimensions(scene);

  PostSetSceneInitDemo(scene);
}


