
#include "NVAssist.h"
#include <NVDemo/NVConstantParams.h>
#include <NVDemo/NVShape.h>
#include <NVDemo/NVShader.h>
#include <NVDemo/NVRenderPass.h>


int snap64(int x)
{
  int y = ((x+32)/64)*64;
  if (y<64) y=64;
  return y;
}
NVec2i snap64(NVec2i  x)
{
  NVec2i y = NVec2i( ((x.x+31)/64)*64, ((x.y+7)/16)*16 );
  if (y.x<64) y.x=64;
  if (y.y<64) y.y=64;
  return y;
}

int RoundUpTo4(int x) {
  return (x + 3) & 0xFFFFFFFC;
}

int RoundUpTo16(int x) {
  return (x + 15) & 0xFFFFFFF0;
}

void RemoveShapeFromGroup(NVShape *s, NVShapeGroup *sg) 
{
  int N = sg->GetChildren().size();
  for (int i=0; i<N; i++) {
    if (sg->GetChildren()[i] == s) {
      sg->RemoveChildren(i);
      return;
    }
  }
  assert(0);
}

void SetUint(NVConstantBuffer* cb, const char* szParamName, uint a, uint b, uint c, uint d) {
  NVUintParam* p = dynamic_cast<NVUintParam*>(cb->FindParam(szParamName));
  assert(p);
  p->SetValue( NPnt4u(a,b,c,d) );
}
void SetFloat(NVConstantBuffer* cb, const char* szParamName, float a, float b, float c, float d) {
  NVFloatParam* p = dynamic_cast<NVFloatParam*>(cb->FindParam(szParamName));
  assert(p);
  p->SetValue( NPnt4f(a,b,c,d) );
}

void
SetGlobalParam(const std::string &shaderName,
	       const std::string &paramName,
	       float minValue,
	       float maxValue,
	       float value){

  NVShader *shader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  if (shader != NULL){
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(shader->FindParam(paramName));
    if (param != NULL){
      float v = minValue + value * (maxValue-minValue);
      param->SetValue(NPnt4f(v, v, v, v));
    }
  }
}

void
SetGlobalParam(const std::string &shaderName,
	       const std::string &paramName,
	       const NPnt4f &minValue,
	       const NPnt4f &maxValue,
	       float value){

  NVShader *shader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  if (shader != NULL){
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(shader->FindParam(paramName));
    if (param != NULL){
      NPnt4f v (minValue.x + value * (maxValue.x-minValue.x),
		minValue.y + value * (maxValue.y-minValue.y),
		minValue.z + value * (maxValue.z-minValue.z),
		minValue.w + value * (maxValue.w-minValue.w));
      param->SetValue(v);
    }
  }
}

void
SetGlobalParam(const std::string &shaderName,
	       const std::string &paramName,
	       const NPnt4f &value){
  NVShader *shader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  if (shader != NULL){
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(shader->FindParam(paramName));
    if (param != NULL)
      param->SetValue(value);
  }
}

NVShape* FindShape(const char *szShapeName) {
  NVShape* s = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName(szShapeName));
  assert(s);
  return s;
}

void
ToggleShape(const std::string &shapeName){
  NVShape* shape = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName(shapeName));
  assert(shape);
  shape->SetCull(shape->GetCull() == NV_CULL_NEVER ? NV_CULL_ALWAYS : NV_CULL_NEVER);
}

void
ToggleShapeGroup(const std::string &shapeGroupName){
  NVShapeGroup* shapeGroup = dynamic_cast<NVShapeGroup*>(NVNAMESPACE->GetByName(shapeGroupName));
  assert(shapeGroup);
  const std::vector<NVBaseShape*> shapes = (shapeGroup->GetChildren());
  for (uint i = 0; i < shapes.size(); i++)
  {
    ToggleShape(shapes[i]->GetName());
  }
}

void
EnableRenderPass(const std::string &renderPassName, bool enable){
  NVRenderPass *rp = dynamic_cast<NVRenderPass*>(NVNAMESPACE->GetByName(renderPassName));
  assert(rp);
  rp->SetActive(enable);
}

void
ToggleRenderPass(const std::string &renderPassName){
  NVRenderPass* renderpassToToggle = (NVRenderPass*)(NVNAMESPACE->GetByName(renderPassName, false));
  assert(renderpassToToggle);
  assert(ISA(renderpassToToggle, NVRenderPass*));
  renderpassToToggle->SetActive(!renderpassToToggle->GetActive());
}

NVRenderPass*
GetRenderPass(const std::string &renderPassName){
  NVRenderPass* pass = (NVRenderPass*)(NVNAMESPACE->GetByName(renderPassName, false));
  assert(pass);
  assert(ISA(pass, NVRenderPass*));
  return pass;
}

void SetTextureOnShader(const std::string &shaderName, const std::string &textureParamParamName, const std::string &textureName)
{
  NVShader *s = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  assert(s);
  NVTextureParam *p = dynamic_cast<NVTextureParam*>(s->FindParam(textureParamParamName));
  assert(p);
  NVTexture* t = dynamic_cast<NVTexture*>(NVNAMESPACE->GetByName(textureName));
  assert(t);
  p->SetTexture(t);
}

void SetFloatParamOnShader(std::string shaderName, std::string paramName, NPnt4f value)
{
  NVShader* s = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName)); 
  assert(s);
  NVFloatParam* fp = dynamic_cast<NVFloatParam*>(s->FindParam(paramName)); 
  assert(fp);
  fp->SetValue( value );
}



float EulerFitAToB(float a, float b)
{
  // return an angle that is on the same place on the unit circle as 'a'
  // but is within +- 3.14 radians of 'b'.
  a /= 6.2831854f;
  b /= 6.2831854f;
  float diff = (b-a);
  if (diff >= 0)
    a += (int)(diff + 0.5f);
  else 
    a -= (int)(-diff + 0.5f);
  return a * 6.2831854f;
} 

NVec3f EulerFitAToB(NVec3f a, NVec3f b)
{
  NVec3f ret;
  ret.x = EulerFitAToB(a.x, b.x);
  ret.y = EulerFitAToB(a.y, b.y);
  ret.z = EulerFitAToB(a.z, b.z);
  return ret;
}

void
SetGlobalParam(const std::string &shaderName,
	       const std::string &paramName,
	       float value)
{
  NVShader *shader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  if (shader != NULL){
    NVFloatParam *param = dynamic_cast<NVFloatParam*>(shader->FindParam(paramName));
    if (param != NULL){
      param->SetValue(NPnt4f(value, value, value, value));
    }
  }
}

void SetTextureOnShape(const std::string &shapeName, const std::string &textureParamParamName, const std::string &textureName)
{
  NVShape *s = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName(shapeName));
  assert(s);
  NVTextureParam *p = dynamic_cast<NVTextureParam*>(s->GetFragmentParam(textureParamParamName));
  assert(p);
  NVTexture* t = dynamic_cast<NVTexture*>(NVNAMESPACE->GetByName(textureName));
  assert(t);
  p->SetTexture(t);
}


std::string ToggleAandB(std::string nameToToggle)
{
  // Changes "somethingA" to "somethingB" and vice-versa
  
  char lastCharacter = nameToToggle[nameToToggle.length()-1];
  if (lastCharacter == 'A')
    return nameToToggle.substr(0, nameToToggle.length()-1) + 'B';
  else if (lastCharacter == 'B')
    return nameToToggle.substr(0, nameToToggle.length()-1) + 'A';
  return nameToToggle;
}

void ToggleTextureParam(std::string shaderName, std::string textureParamName)
{
  NVShader* theShader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  assert (theShader);
  NVTextureParam* theTextureParam = dynamic_cast<NVTextureParam*>(theShader->FindParam(textureParamName));
  assert (theTextureParam);
  std::string theTextureName = ToggleAandB(theTextureParam->GetTexture()->GetName());
  NVTexture* theTexture = dynamic_cast<NVTexture*>(NVNAMESPACE->GetByName(theTextureName));
  assert (theTexture);
  theTextureParam->SetTexture(theTexture);
}

void ToggleBufferParam(std::string shaderName, std::string bufferParamName)
{
  NVShader* theShader = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaderName));
  assert (theShader);
  NVBufferParam* p = dynamic_cast<NVBufferParam*>(theShader->FindParam(bufferParamName));
  assert (p);
  //std::string newName = ToggleAandB(p->GetBuffer()->GetName());
  //NVDataNode* newBuffer = dynamic_cast<NVDataNode*>(NVNAMESPACE->GetByName(newName));
  //assert(newBuffer);
  //p->SetBuffer(newBuffer);
  std::string newName = ToggleAandB(p->GetGeometry()->GetName());
  NVGeometry* newGeom = dynamic_cast<NVGeometry*>(NVNAMESPACE->GetByName(newName));
  assert(newGeom);
  p->SetGeometry(newGeom);

}



bool BBoxIsOutsideViewFrustum(NPnt3f pt1, NPnt3f pt2, NVBaseCamera* cam) 
{
  const NPlanef planes[6] = { cam->GetNearPlane  (),
                              cam->GetFarPlane   (),
                              cam->GetLeftPlane  (),
                              cam->GetRightPlane (),
                              cam->GetTopPlane   (),
                              cam->GetBottomPlane() 
                            };
  NPnt3f pts[8] = { NPnt3f(pt1.x, pt1.y, pt1.z), 
                    NPnt3f(pt1.x, pt1.y, pt2.z), 
                    NPnt3f(pt1.x, pt2.y, pt1.z), 
                    NPnt3f(pt1.x, pt2.y, pt2.z), 
                    NPnt3f(pt2.x, pt1.y, pt1.z), 
                    NPnt3f(pt2.x, pt1.y, pt2.z), 
                    NPnt3f(pt2.x, pt2.y, pt1.z), 
                    NPnt3f(pt2.x, pt2.y, pt2.z) 
                  };

  for (int p=0; p<6; p++) {
    bool bAllOutside = true;
    for (int v=0; v<8; v++) {
      float d = planes[p].distance(pts[v]);
      if (d > 0) {
        bAllOutside = false;
        break;
      }
    }
    if (bAllOutside)
      return true;
  }

  return false;
}


void ToggleInputGeometryAandB(const std::string shapeName)
{
  NVShape* theShape = dynamic_cast<NVShape*>(NVNAMESPACE->GetByName(shapeName));
  assert (theShape);
  NVGeometry* theGeometry = dynamic_cast<NVGeometry*>(NVNAMESPACE->GetByName( ToggleAandB(theShape->GetGeometry()->GetName()) ));
  theShape->SetGeometry(theGeometry);
}

void ToggleStreamOutTargetAandB(const std::string passName)
{
  NVRenderPass* pass = GetRenderPass(passName);
  std::vector<NVDataNode*> SOTargets = pass->GetStreamTargets();
  for (uint i=0; i<SOTargets.size(); i++) {
    std::string newSOTargetName = ToggleAandB(SOTargets[i]->GetName());
    NVDataNode* newSOTarget = dynamic_cast<NVDataNode*>(NVNAMESPACE->GetByName( newSOTargetName ));
    assert (newSOTarget);
    SOTargets[i] = newSOTarget;
  }
  pass->SetStreamTargets(SOTargets);
}

void ExactSleep(float ms)
{
  double end_time = NGetTime() + (double)(ms*0.001);
  int its = 0;
  while (NGetTime() < end_time)
  {
    its++;
  };
}
