#ifndef _NVASSIST_H_
#define _NVASSIST_H_

#include <NVDemo/NVViewer.h>

class NVBaseCamera;
class NVShape;
class NVShapeGroup;
class NVShader;
class NVRenderPass;
class NVFloatParam;
class NVUintParam;
class NVConstantBuffer;

#define SAFE_DELETE(x) { if (x) { delete x; x = NULL; } }

inline float lerp(float x,float y,float z) { return x*(1.0f-z) + y*z; }
inline NVec3f lerp(NVec3f a,NVec3f b,float t) { float t2 = 1.0f-t; return NVec3f(a.x*t2+b.x*t, a.y*t2+b.y*t, a.z*t2+b.z*t); }
inline NPnt3f lerp(NPnt3f a,NPnt3f b,float t) { float t2 = 1.0f-t; return NPnt3f(a.x*t2+b.x*t, a.y*t2+b.y*t, a.z*t2+b.z*t); }
inline NVec4f lerp(NVec4f a,NVec4f b,float t) { float t2 = 1.0f-t; return NVec4f(a.x*t2+b.x*t, a.y*t2+b.y*t, a.z*t2+b.z*t, a.w*t2+b.w*t); }
inline NPnt4f lerp(NPnt4f a,NPnt4f b,float t) { float t2 = 1.0f-t; return NPnt4f(a.x*t2+b.x*t, a.y*t2+b.y*t, a.z*t2+b.z*t, a.w*t2+b.w*t); }

#ifndef FRAND
  #define FRAND ((rand()%19734)/19733.0f)
#endif
#define FRAND2 (0.5f*(FRAND + FRAND))
#define FRAND3 (0.333333f*(FRAND + FRAND + FRAND))
#define RANDSIGN ((FRAND < 0.5f) ? 1 : -1)

#define IS_UPPER(c) ((c) >= 'A' && (c) <= 'Z')
#define IS_LOWER(c) ((c) >= 'a' && (c) <= 'z')
#define TO_UPPER(c) (IS_UPPER(c) ? (c) : ((c) - ('a' - 'A')))
#define TO_LOWER(c) (IS_LOWER(c) ? (c) : ((c) + ('a' - 'A')))

int snap64(int x);
NVec2i snap64(NVec2i  x);
int RoundUpTo4(int x);
int RoundUpTo16(int x);

float EulerFitAToB(float a, float b);  // returns an angle that is on the same place on the unit circle as 'a' but is within +- 3.14 radians of 'b'.
NVec3f EulerFitAToB(NVec3f a, NVec3f b);

bool BBoxIsOutsideViewFrustum(NPnt3f pt1, NPnt3f pt2, NVBaseCamera* cam);

void ExactSleep(float ms);

//---------------------------
// NVDEMO STUFF:

void RemoveShapeFromGroup(NVShape *s, NVShapeGroup *sg);

void SetUint(NVConstantBuffer* cb, const char* szParamName, uint a, uint b, uint c, uint d);
void SetFloat(NVConstantBuffer* cb, const char* szParamName, float a, float b, float c, float d);
void SetGlobalParam(const std::string &shaderName, const std::string &paramName,	       
                    float minValue, float maxValue, float value);
void SetGlobalParam(const std::string &shaderName, const std::string &paramName,
                    const NPnt4f &minValue, const NPnt4f &maxValue, float value);
void SetGlobalParam(const std::string &shaderName, const std::string &paramName, float value);
void SetGlobalParam(const std::string &shaderName, const std::string &paramName, const NPnt4f &value);
void SetFloatParamOnShader(std::string shaderName, std::string paramName, NPnt4f value);

void SetTextureOnShader(const std::string &shaderName, const std::string &textureParamParamName, const std::string &textureName);
void SetTextureOnShape(const std::string &shapeName, const std::string &textureParamParamName, const std::string &textureName);

NVShape* FindShape(const char *szShapeName);
void ToggleShape(const std::string &shapeName);
void ToggleShapeGroup(const std::string &shapeGroupName);

NVRenderPass* GetRenderPass(const std::string &renderPassName);
void EnableRenderPass(const std::string &renderPassName, bool enable);
void ToggleRenderPass(const std::string &renderPassName);

std::string ToggleAandB(std::string nameToToggle);
void        ToggleTextureParam(std::string shaderName, std::string textureParamName);
void        ToggleBufferParam(std::string shaderName, std::string bufferParamName);
void        ToggleInputGeometryAandB(const std::string shapeName);
void        ToggleStreamOutTargetAandB(const std::string passName);







#endif _NVASSIST_H_