#ifndef _NVDEMO_RELOAD_SHADERS_
#define _NVDEMO_RELOAD_SHADERS_

#include <NVDemo/NVScene.h>

// this is pretty slow:
void ReloadAllShaders(NVScene* scene);

// this is much faster, if 'count' is less than 8 or so.
// 'shaders_to_reload' is an array of the names of the shaders 
//     ("hair_f", "fishboy_v", etc.)
// also note that this only works on shaders that get used on
//     NVShape-derived objects.  (it won't work, for example,
//     on a shader used for a menu item (NVMenuItem-derived),
//     because it's not smart enough to hit all its shader params
//     and flush their bindcaches.)
void ReloadSomeShaders(NVScene *scene, std::string *shaders_to_reload, int count);
void ReloadSomeShaders(NVScene *scene, std::vector<std::string> &shaders_to_reload);

/* 
example usage: 
  in your demo, in OnKeyDown, add cases like these:

  case 133: // F5
    ReloadAllShaders( m_viewer->GetScene() );
    break;
  
  case 8:   // backspace
    {
      std::string shaders_to_reload[] = {
        "hair_v",   // PUT LIST OF SHADERS 
        "hair_f",   //  YOU WANT TO RELOAD 
        "..."       //   HERE
      };
      ReloadSomeShaders( m_viewer->GetScene(), shaders_to_reload, sizeof(shaders_to_reload)/sizeof(std::string) );
    }
    break;

*/ 



#endif



