/* 
 *
 * COPYRIGHT NVIDIA CORPORATION 2002. ALL RIGHTS RESERVED.  
 * BY ACCESSING OR USING THIS SOFTWARE, YOU AGREE TO: 
 *
 *  1) ACKNOWLEDGE NVIDIA'S EXCLUSIVE OWNERSHIP OF ALL RIGHTS 
 *     IN AND TO THE SOFTWARE; 
 *
 *  2) NOT MAKE OR DISTRIBUTE COPIES OF THE SOFTWARE WITHOUT 
 *     INCLUDING THIS NOTICE AND AGREEMENT; 
 *
 *  3) ACKNOWLEDGE THAT TO THE MAXIMUM EXTENT PERMITTED BY 
 *     APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS* AND 
 *     THAT NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, 
 *     EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED
 *     TO, IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS 
 *     FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY 
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES 
 * WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS 
 * OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
 * INFORMATION, OR ANY OTHER PECUNIARY LOSS), INCLUDING ATTORNEYS'
 * FEES, RELATING TO THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 */

#ifndef _MMM_SUPPORT_H_
#define _MMM_SUPPORT_H_

#include <NCommon/NCommon.h>
#include <NVDemo/NVGeometry.h>
#include <NMath/NNoise.h>
#include <NMath/NVec.h>
#include <NMath/NMat.h>

#ifndef FRAND
  #define FRAND ((rand()%19734)/19733.0f)
#endif

float AdjustRateToFPS(float per_frame_decay_rate_at_fps1, float fps1, float actual_fps);

inline NPnt3f VecToPnt(const NVec3f &pt) { return NPnt3f(pt.x, pt.y, pt.z); }
inline NVec3f PntToVec(const NPnt3f &pt) { return NVec3f(pt.x, pt.y, pt.z); }
inline NPnt2f VecToPnt(const NVec2f &pt) { return NPnt2f(pt.x, pt.y); }
inline NVec2f PntToVec(const NPnt2f &pt) { return NVec2f(pt.x, pt.y); }

NVec3f TransformProjective   (const NVec3f &pt, const NMatf &m);  // does divide by w
NPnt3f TransformProjective   (const NPnt3f &pt, const NMatf &m);  // does divide by w
NVec3f TransformNonProjective(const NVec3f &pt, const NMatf &m);  // skips divide by w (assumes w' == 1)
NPnt3f TransformNonProjective(const NPnt3f &pt, const NMatf &m);  // skips divide by w (assumes w' == 1)
NVec3f Transform3x3          (const NVec3f &pt, const NMatf &m);  // no translate

inline float saturate(float f) { if (f>1) return 1; if (f<0) return 0; return f; }
inline float smooth(float value, int its=1) { value = min(1, max(0, value)); for (int i=0; i<its; i++) value = value*value*(3 - 2*value); return value; }

int GetIntFromDistribution(int* distribution, int nIndices);
int RandFromProb(float* cum_prob_array, int array_size);

void ScaleVerts(NVBaseMesh *baseMesh, const NVec3f &scale);

NVec3f GetMatrixTranslation(const NMatf &m);

void SetScrollLock(bool bNewState);

int  NVGetIntFromReg(const char* szDemoName, const char* szValueName, int defaultValue);
bool NVWriteIntToReg(const char* szDemoName, const char* szValueName, int value);
void NVGetStringFromReg(const char* szDemoName, const char* szValueName, char *szBuf, int nBufSize, const char* szDefault);
bool NVWriteStringToReg(const char* szDemoName, const char* szValueName, const char* szStringToWrite);
bool NVGetBinaryFromReg(const char* szDemoName, const char* szValueName, void *dest, unsigned int destBufSize);
bool NVWriteBinaryToReg(const char* szDemoName, const char* szValueName, void *data, unsigned int nBytes);
bool NVKeyExists(const char* szDemoName, const char* szValueName);
bool NVDeleteKey(const char* szDemoName, const char* szValueName);

class CFpsMonitor {
public:
  CFpsMonitor(int nFramesToTrack = 5400);  // 5400 gives you 3m:00s damping @ 30 fps
  ~CFpsMonitor();
  void Reset();
  float Tick(float unpausable_delta_t);
    // note: if you want your fps to keep updating while demo is paused, keep calling Tick().
    //       if you want your fps to freeze, just don't call Tick().  It'll filter out the long pause.
  inline float GetDampledFps() { return m_fDampedFps; };
  inline int   GetLogSize() { return m_nLogSize; };
  inline int   GetNumTicksSinceReset() { return m_nFramesSinceReset; };
protected:
  int    m_nLogSize;
  int    m_nFramesSinceReset;
  float* m_fFpsLog;
  float  m_fFpsLogSum;
  float  m_fDampedFps;
};


#endif